# sage_setup: distribution = sagemath-repl
"""
This file (./sol/linsolve_doctest.sage) was *autogenerated* from ./sol/linsolve.tex,
with sagetex.sty version 2011/05/27 v2.3.1.
It contains the contents of all the sageexample environments from this file.
You should be able to doctest this file with:
sage -t ./sol/linsolve_doctest.sage
It is always safe to delete this file; it is not used in typesetting your
document.

Sage example in ./sol/linsolve.tex, line 72::

  sage: m = random_matrix(RDF,4)
  sage: a = transpose(m)*m
  sage: c = a.cholesky()
  sage: U,S,V = c.SVD()
  sage: X = U*S*transpose(U)

Sage example in ./sol/linsolve.tex, line 84::

  sage: M = (X*X-a)
  sage: all(abs(M[i,j]) < 10^-13
  ....:     for i in range(4) for j in range(4) )
  True
"""
